from functools import wraps
from hashlib import sha256
from hmac import new as hmac
from time import time
from typing import Any, Callable, Union, cast

from .types import Endpoint, RequestProtocol, StringProvider


class SlackHeaders(object):
    X_SLACK_SIGNATURE = "X-Slack-Signature"
    X_SLACK_REQUEST_TIMESTAMP = "X-Slack-Request-Timestamp"


class InvalidSignatureError(ValueError):
    pass


class ExpiredTimestampError(ValueError):
    pass


def compute_slack_signature(
    request: RequestProtocol,
    signing_secret: str,
    version: str,
) -> str:
    """Compute a Slack signature from the given request."""
    timestamp = request.headers.get(SlackHeaders.X_SLACK_REQUEST_TIMESTAMP)
    if abs(time() - float(cast(int, timestamp))) > 60 * 5:
        raise ExpiredTimestampError()
    body = request.get_data().decode("utf-8")
    message = f"{version}:{timestamp}:{body}"
    signature = hmac(
        signing_secret.encode("utf-8"),
        msg=message.encode("utf-8"),
        digestmod=sha256,
    )
    return f"{version}={signature.hexdigest()}"


def verify_slack_signature(
    signing_secret: Union[str, StringProvider],
    version: str = "v0",
) -> Callable[[Endpoint], Endpoint]:
    """
    Decorator that acts as middleware to ensure that an
    incoming HTTP request is a valid issued Slack request.

    See https://api.slack.com/authentication/verifying-requests-from-slack
    """

    def wrapper(endpoint: Endpoint) -> Endpoint:
        @wraps(endpoint)
        def middleware(request: RequestProtocol) -> Any:
            if isinstance(signing_secret, str):
                secret = signing_secret
            else:
                secret = signing_secret()
            expected = request.headers.get(SlackHeaders.X_SLACK_SIGNATURE)
            actual = compute_slack_signature(
                request,
                secret,
                version,
            )
            if actual != expected:
                raise InvalidSignatureError()
            return endpoint(request)

        return middleware

    return wrapper



from typing import Any, Callable, Coroutine, Mapping, Protocol

from .blocks import Blocks


class RequestProtocol(Protocol):
    """A headers provider interface for duck typing."""

    @property
    def headers(self) -> Mapping[str, Any]:
        ...

    def get_data(self) -> bytes:
        ...


Endpoint = Callable[[RequestProtocol], Any]

AsyncWebhook = Callable[[Blocks], Coroutine[Any, Any, None]]
Webhook = Callable[[Blocks], None]

StringProvider = Callable[..., str]



from typing import List

from pydantic import BaseModel


class BlockAction(BaseModel):
    type: str
    block_id: str
    action_id: str
    value: str
    action_ts: str


class BlockInteraction(BaseModel):
    type: str
    actions: List[BlockAction]
    response_url: str


class InteractionDeleteResponse(BaseModel):
    delete_original: bool = True


class InteractionReplaceResponse(BaseModel):
    replace_original: bool = True
    text: str



from functools import lru_cache

from httpx import AsyncClient, Client

from .blocks import Blocks
from .types import AsyncWebhook, Webhook


@lru_cache(maxsize=10)
def SlackWebhook(url: str) -> Webhook:
    client = Client()

    def webhook(blocks: Blocks) -> None:
        response = client.post(url, json=blocks.dict())
        response.raise_for_status()

    return webhook


@lru_cache(maxsize=10)
def AsyncSlackWebhook(url: str) -> AsyncWebhook:
    client = AsyncClient()

    async def webhook(blocks: Blocks) -> None:
        response = await client.post(url, json=blocks.dict())
        response.raise_for_status()

    return webhook
